open Util

open Kernel

open Data.Type

let phrase_sz = 17

let prompt = "# "

let show_prompt () =  
  let nothing_read = Threadref.make true in
  let f () =
    Thread.delay 0.05;
    if (Threadref.get nothing_read) then
      begin
        print_string prompt;
        flush stdout;
      end in
  ignore (Thread.create f ());
  nothing_read

let rec read_loop nothing_read buffer =
  let eof = ref false in
  try
    let s =
      try
        read_line ()
      with
        | End_of_file ->
            print_newline ();
            exit 0 in
    Threadref.set nothing_read false;
    Buffer.add_string buffer s;
    let get i =
      try Some (Buffer.nth buffer i)
      with Invalid_argument _ ->
        eof := true;
        None in
    let stream = Stream.from get in
    let (package, ast) = Reader.parse_stream "" stream in
    begin
      match package with
        | None -> ()
        | Some _ -> failwith "No package declaration in toplevel"
    end;
    ast
  with
    | e ->
        if !eof then
          read_loop nothing_read buffer
        else
          raise e
and read_phrase () =
  let nothing_read = show_prompt () in
  let buffer = Buffer.create phrase_sz in
  read_loop nothing_read buffer

let execute (global_vars, global_array) dollar ast =
  Compile.reset ();
  let mk v = { Pos.v = v; range = Pos.dummy_range } in
  let old_env = !Ast.env in
  try
    Compile.file None "./" "silcc/top.lcc";
    let ast = mk (Ast.Load_env (mk (Ast.Concat (ast, mk Ast.Save_env)))) in
    Compile.of_ast false "silcc.top" ast;
    let opt = Compile.all () in
    let (am, global_vars') = Am.of_opt opt global_vars in
    let global_array' = Array.make global_vars'.Am.count Interpret.dummy in
    Array.blit global_array 0 global_array' 0 (Array.length global_array);
    Interpret.execute am global_array';
    let (env_count, env_list) = !Ast.env in
    let sy =
      Interpret.V.sync dollar
        (Sgn.get (Ast.env_name env_count) (List.length env_list)) in
    if sy.Interpret.Sync.signals = [] then
      Ast.env := old_env;
    (global_vars', global_array')
  with e ->
    Compile.forget_new_compiled_modules ();
    Ast.env := old_env;
    raise e

let rec top env dollar =
  let env =
    try
      let ast = read_phrase () in
      execute env dollar ast;
    with
      | Err.E (pos, msg) ->
          Err.print pos msg;
          env
      | e ->
          Format.eprintf "Error. %s.@." (Printexc.to_string e);
          env in
  flush stderr;
  top env dollar

let run () =
  let global_vars = { Am.count = 1; map = Var.Map.add "$" 0 Var.Map.empty } in
  let dollar = Interpret.V.fresh () in
  let global_array = [| Interpret.var_ty.inj dollar |] in
  top (global_vars, global_array) dollar
